import operator
import fractions
from linkedlist import *
from symbol import *
import sys

class Sexp(object):
    """ 
        It represents the nodes of a tree which are procedures.
        Each procedure contains an id a.k.a name and arguments.

        As scheme code is evaluated top down there is a main procedure whose id is 'main'.

        The main procedure is evaluated first. It is a virtual procedure in the sense
        that the user doesn't need to implement, it is automatically built by the scan.

        The arguments for the main procedure are the pieces of the scheme source code.

        The main procedure will just evaluate all of its arguments, 
        providing so a scope for the entire program.
    """

    def __call__(self, scope):
        return self.id(scope)(scope, self.args)


    def mapid(self, scope):
        return self.id(scope)

    def __init__(self, *args):
        # Otherwise it would fail on '().
        # It might be interesting of having just self.args = args.
        try:
            self.id   = args[0]
            self.args = args[1:]
        except IndexError:
            pass

        self.data = args

    def generate_linkedlist(self):
        head = Nil()
        for ind in reversed(self.data):
            head = cons(ind.generate_linkedlist(), head)
        return head

    def __str__(self):
        data = ['(']
        data.append(str(self.id))
        for ind in self.args:
            data.append(str(ind))
        data.append(')')

        return ' '.join(data)

    def exec_args(self, scope):
        """
        This function evaluates parcially its sexp arguments.
        It returns a new sexp with its arguments parcially evaluated.
        """
        sexp = Sexp(self.id, *eval_args(scope, self.args))
        return sexp

    def evalrec(self, scope):
        return self.id.evalrec(scope).evalrec(scope, self.args)
    
class Main(Sexp):
    def __call__(self, scope):
        for ind in self.args:
            ind(scope)

    def __init__(self, *args):
        Sexp.__init__(self, Sym('main'), *args)

    def __str__(self):
        data = []
        for ind in self.args:
            data.append(str(ind))
        return ' '.join(data)        

    def exec_args(self, scope):
        pass

class UnknownSymbol(Exception):
    def __init__(self, *args):
        self.args = args

    def __repr__(self):
        return 'Unknown symbol: ', self.args[0]


def readtype(id):
    try:
        value = int(id)
    except:
        pass
    else:
        return Int(value)

    try:
        value = float(id)
    except:
        pass
    else:
        return Float(value)

    return Sym(id)


class Sym(Sexp):
    def __call__(self, scope):
        try:
            value = scope[self.id]
        except KeyError:
            raise UnknownSymbol(repr(self.id))
        else:
            return value

    def mapid(self, scope):
        return self

    def __init__(self, id):
        Sexp.__init__(self, id)

    def __str__(self):
        return str(self.id)

    def exec_args(self, scope):
        """ 
        As it is a symbol the evaluation of its argument should be
        the evaluation of itself inside the scope. 
        Otherwise it would break the Let class.
        """
        return self(scope)

    def generate_linkedlist(self):
        return self.id


    def evalrec(self, scope):
        return self.__call__(scope)


class Str(Sym):
    def __call__(self, scope):
        return self.id

    def mapid(self, scope):
        return self

    def __init__(self, id):
        Sym.__init__(self, id)

    def __str__(self):
        return '"%s"' % self.id


class Int(Sym):
    def __call__(self, scope):
        return self.id

    def mapid(self, scope):
        return self

    def __init__(self, id):
        Sym.__init__(self, id)


class Float(Sym):
    def __call__(self, scope):
        return self.id

    def mapid(self, scope):
        return self

    def __init__(self, id):
        Sym.__init__(self, id)


class Tail(object):
    def __init__(self):
        pass

    def evalrec(self, scope, args):
        pass

class NonTail(object):
    def __init__(self):
        pass

    def evalrec(self, scope, args):
        return self.__call__(scope, args)


class Define(NonTail):
    def __call__(self, scope, args):
        # It makes impossible to redefine functions. 
        # once they are defined they are evaluated to Op
        # so it cant update the function name in the scope.

        id          = str(args[0].id)
        local       = [str(ind) for ind in args[0].args]
        scope[id]   = Op(scope, id, local, args[1:-1], args[-1])
        # The local scope var names is named local.
        # (define (name x)
        #         (+ 1 x))
        #
        # name  = 'name'
        # local = ['x']

        return scope[id]



class Op(object):
    def __call__(self, scope, args):
        # Just to save some typing and for cleaness.
        # Holy fuck. It took me a long time to perceive
        # that.

        # (define (sum-n-first-numbers n) (if (> n 0) 
        #                            (+ (sum-n-first-numbers (- n 1)) n) 0))
        # (display (sum-n-first-numbers 4))

        # Would evail to 0 if i don't copy the self.scope.

        env   = self.env.copy()
        local = self.local
        last  = self.last
        code  = self.code

        env.update(scope)

        # I need to update self.env with the new scope.
        # Otherwise things like this would fail.

        # (define (ae x)
        #         (+ 1 x))
        # (define (bind u a) (ae u))

        # I must evaluate the arguments in the extern scope.
        # Otherwise i will have things like.

        #  (define (foo x) (+ x a))
        #  (define (bar a) (foo a))
        #  (display (bar 2))
        #  Result: 4

        update_scope(env, local, args)

        # In case of an ethernal loop.
        if last.mapid(env) == self:
            while 1:
                for ind in code:
                    ind(env)
    
                    update_scope(env, local, last.args)
    

        while 1:
        # This has to be inside the while 1:
        # otherwise things like.

        # (define (foo x)
        #         (display x)
        #         (if (< x 10) (foo (+ x 1)) x))
        # would fail to print 'x' at each iteration.
            for ind in code:
                ind(env)

                
            sexp = last.evalrec(env)
            if not isinstance(sexp, Sexp): return sexp
            if sexp.mapid(env) == self:
                update_scope(env, local, sexp.args)
            else: return sexp(env)


    def __init__(self, scope, id, local, code, last):
        self.id     = id
        self.local  = local
        self.code   = code
        self.last   = last

        # It requires holding an internal copy
        # of the scope. Otherwise we can't have stuff
        # like functions being returned from functions
        # and still saving the previous scope.


        # (define (foo x) 
        #         (define (bar) (x))
        #         bar)

        # (display ((foo 10))

        # Things like this would fail.

        self.env    = dict()
        self.env.update(scope)

    def evalrec(self, scope, args):
        return self.__call__(scope, args)


class Lambda(NonTail):
    def __call__(self, scope, args):
        local = [str(ind) for ind in args[0].data]
        code  = args[1]

        env = dict()
        env.update(scope)

        def run(scope, args, env=env, local=local, code=code):
            env.update(scope)
            update_scope(env, local, args)
            return code(env)
    
        return run


class Map(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        head = sons(Nil(), a(scope, (b.value, )))
        b = b.back

        last = head
        while not isinstance(b, Nil):
            last = sons(last, a(scope, (b.value,))) 
            b = b.back
        return head


class Reduce(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()
        b = args.next()
        c = args.next()

        # a = args[0]
        # b = args[1]
        # c = args[2]

        if isinstance(c, Nil):
            return b

        if isinstance(c.back, Nil):
            return c.value

        # I'm passing ints to functions like +.
        # (display (reduce + 0 \'(1 2 3))) breaks.


        result = a(scope, (c.value, c.back.value))
        c = c.back.back

        while not isinstance(c, Nil):
            result = a(scope, (result, c.value))
            c = c.back
        return result
        

class Filter(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]


        if isinstance(b, Nil):
            return Nil()

        if isinstance(b.back, Nil):
            if a(scope, (b.value, )):
                return cons(b.value, Nil())
            else:
                return Nil()

        head = None
        while not isinstance(b, Nil):
            if a(scope, (b.value, )):
                head = sons(Nil(), b.value)
                b = b.back
                break
            b = b.back
        else:
            return Nil()
        last = head
        while not isinstance(b, Nil):
            if a(scope, (b.value,)):
                last = sons(last, b.value) 
            b = b.back
        return head



class StringNumber(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()
        # a = args[0]

        try:
            value = int(a)
        except:
            pass
        else:
            return value

        value = float(a)

        return value


class StringDrop(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]


        return a[b:]


class StringTake(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        return a[:b]

class StringLength(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()

        # a = args[0]

        return len(a)


class Append(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        return append(a, b)

class Length(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()

        # a = args[0]

        return length(a)


def eval_args(scope, args):
    args = (ind(scope) if isinstance(ind, Sexp) else ind for ind in args)
    return args

def update_scope(scope, local, args):
    done = eval_args(scope, args)
    lookup = zip(local, done)
    scope.update(lookup)

class Sum(NonTail):
    """ 
        Builtin functions are builtin procedures
        whose job is evaluating sexps and performing
        its task over the sexps results.
    """
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.add, args)

    def __init__(self):
        pass

class Sub(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.sub, args)
        pass

    def __init__(self):
        pass

class Div(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.div, map(fractions.Fraction, args))

    def __init__(self):
        pass

class Quotient(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        return int(a/b)


class Mul(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.mul, args)

    def __init__(self):
        pass

class Expt(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        return a ** b

class IsZero(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()

        # a = args[0]
        return a == 0

class Cond(Tail):
    def __call__(self, scope, args):
        for ind in args:
            a = ind.data[0](scope)
            if a:
                return ind.data[1](scope)

    def evalrec(self, scope, args):
        for ind in args:
            a = ind.data[0](scope)
            if a:
                return parcial(scope, ind.data[1])

class Let(Tail):
    def __call__(self, scope, args):
        """ 
            I don't need to worry of dirtying the scope.
            scheme is a functional lang.
        """

        env = dict()
        env.update(scope)

        for ind in args[0].data:
            env[str(ind.data[0])] = ind.data[1](scope)

        # It will break if
        # (let ((a 1))
        #      a)
        # Sym should be a Sexp. it will bring cleaness.

        for ind in args[1:-1]:
            ind(env)
            
        # (let ((q 1))
        #      (define (alpha) 
        #              (+ 1 1)))

        # I can just evaluate the arguments to args[-1].id here.
        # So, they will still hold inside the copy of the context.
        # We then will be able to do tail recursion optmization here.
        
        return args[-1](env)


    def evalrec(self, scope, args):
        env = dict()
        env.update(scope)

        for ind in args[0].data:
            env[str(ind.data[0])] = ind.data[1](scope)

        # It will break if
        # (let ((a 1))
        #      a)
        # Sym should be a Sexp. it will bring cleaness.

        for ind in args[1:-1]:
            ind(env)
            
        # (let ((q 1))
        #      (define (alpha) 
        #              (+ 1 1)))

        # I can just evaluate the arguments to args[-1].id here.
        # So, they will still hold inside the copy of the context.
        # We then will be able to do tail recursion optmization here.
        
        # I need this otherwise things like:
        # (define (foo)
        #         (let ((q 1))
        #              (display 'bar')
        #              (foo)))
        # wouldn't be possible as tail recursion.

        if isinstance(args[-1].mapid(env), Op):
            return args[-1].exec_args(env)

        result = args[-1].evalrec(env)
        if isinstance(result, Sexp):
            if isinstance(result.mapid(env), Op):
                return result.exec_args(env)
        return result


class If(Tail):
    def __call__(self, scope, args):
        a = args[0]
        b = args[1]
        c = args[2]
    
        r = a(scope)
    
        if r:
            return b(scope)
        else:
            return c(scope)

    def evalrec(self, scope, args):
        a = args[0]
        b = args[1]
        c = args[2]

        r = a(scope)

        if r:
            return parcial(scope, b)
        else:
            return parcial(scope, c)

def parcial(scope, code):
    while 1:
        # I might need this cause (if (= 0 0) (+ 1 1) 0)
        # would return an int.
        # on the other hand this function is being called only by
        # If, cond, Let which forcely should make code be a Sexp instance.

        if not isinstance(code, Sexp): return code

        if isinstance(code.mapid(scope), (If, Cond, Let)):
            code = code.evalrec(scope)
        elif not isinstance(code.mapid(scope), Op):
            return code(scope)
        else:
            return code

class Or(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return any(args)


class And(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return all(args)


class Gt(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.gt, args)

class Ge(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.ge, args)


class Not(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()

        # a = args[0]
        return not a


class Lt(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.lt, args)

class Le(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        return reduce(operator.le, args)

class Remainder(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        return a % b

    def __init__(self):
        pass

class IsPositive(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()

        # a = args[0]
        return a > 0

class IsInteger(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()

        # a = args[0]
        return int(a) == a 

class Round(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()

        # a = args[0]
        return round(a)


class Sqrt(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()

        # a = args[0]
        return a ** (1/2.0)


class Eq(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        return a == b

    def __init__(self):
        pass

class Display(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        for ind in args:
            sys.stdout.write('%s\n' % ind)
            sys.stdout.flush()
            # print ind

class Quote(Sym):
    def __call__(self, scope):
        return self.id

    def __init__(self, id):
        Sym.__init__(self, id.generate_linkedlist())

    def __str__(self):
        return "%s" % str(self.id)


class Cons(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        b = args.next()

        # a = args[0]
        # b = args[1]

        return cons(a, b)

class Car(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()

        # a = args[0]
        return a.value

class Cdr(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        a = args.next()
        # a = args[0]
        return a.back

class List(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)

        head = Nil()
        for ind in reversed(tuple(args)):
            head = cons(ind, head)

        return head


class Null(NonTail):
    def __call__(self, scope, args):
        args = eval_args(scope, args)
        a = args.next()

        # a = args[0]
        return isinstance(a, Nil)

class Begin(object):
    def __call__(self, scope, args):
        pass

class Letrec(Tail):
    def __call__(self, scope, args):
        pass

class LetAsterisk(Tail):
    def __call__(self, scope, args):
        pass


class Take(NonTail):
    def __call__(self, scope, args):
        pass


class Drop(NonTail):
    def __call__(self, scope, args):
        pass


GLOBAL_SCOPE = {
                   'define' : Define(), 
                   'if': If(),
                   'or': Or(),
                   'and': And(),
                   '>' : Gt(),
                   '<' : Lt(),
                   '<=' : Le(),
                   '>=' : Ge(),
                   '+': Sum(),
                   '-': Sub(),
                   '/': Div(),
                   'quotient': Quotient(),
                   'list' : List(),
                   '*': Mul(),
                   'lambda': Lambda(),
                   'let' : Let(),
                   'map' : Map(),
                   'null?' : Null(),
                   'expt' : Expt(),
                   'cond' : Cond(),
                   'cons' : Cons(),
                   'car' : Car(), 
                   'cdr' : Cdr(),
                   'positive?' :IsPositive(),
                   'integer?' :IsInteger(),
                   'sqrt': Sqrt(),
                   'round': Round(),
                   'zero?' : IsZero(),
                   'else': True,
                   'append': Append(),
                   'remainder': Remainder(),
                   '=': Eq(), 
                   'not' : Not(),
                   '#t': True,
                   '#f': False, 
                   'reduce': Reduce(),
                   'length': Length(),
                   'filter' : Filter(),
                   'string->number': StringNumber(),
                   'string-take': StringTake(),
                   'string-length' : StringLength(),
                   'string-drop' : StringDrop(),
                   'display': Display()
               }
















































